#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

#   Author: Laura Murgatroyd (UKRI-STFC)
#   Author: Edoardo Pasca (UKRI-STFC)

import os
import numpy as np
from PySide2 import QtCore
from datetime import datetime
from PySide2.QtWidgets import QMessageBox
import json
import time
import shutil
import platform
from .io import save_tiff_stack_as_raw

class PrintCallback(object):
    '''Class to handle the emit call when no callback is provided'''
    def emit(self, *args, **kwargs):
        print (args, kwargs)

count = 0
runs_completed = 0

blank_config = '''###############################################################################
#																	
#
#		example dvc process control file		
#
#
###############################################################################

# all lines beginning with a # character are ignored
# some parameters are conditionally required, depending on the setting of other parameters
# for example, if subvol_thresh is off, the threshold description parameters are not required

### file names

reference_filename	{reference_filename}		### reference tomography image volume
correlate_filename	{correlate_filename}		### correlation tomography image volume

point_cloud_filename	{point_cloud_filename}	### file of search point locations
output_filename		{output_filename}		### base name for output files

### description of the image data files, all must be the same size and structure

vol_bit_depth		{vol_bit_depth}			### 8 or 16
vol_hdr_lngth		{vol_hdr_lngth}		### fixed-length header size, may be zero
vol_wide		{vol_wide}			### width in pixels of each slice
vol_high		{vol_high}			### height in pixels of each slice
vol_tall		{vol_tall}			### number of slices in the stack
vol_endian		{vol_endian}        ### big or little endian byte ordering

### parameters defining the subvolumes that will be created at each search point

subvol_geom		{subvol_geom}			### cube, sphere
subvol_size		{subvol_size}			### side length or diameter, in voxels
subvol_npts		{subvol_npts}			### number of points to distribute within the subvol

subvol_thresh		{subvol_thresh}			### on or off, evaluate subvolumes based on threshold
gray_thresh_min		{gray_thresh_min}			### lower limit of a gray threshold range if subvol_thresh is on
gray_thresh_max		{gray_thresh_max}			### upper limit of a gray threshold range if subvol_thresh is on
min_vol_fract		{min_vol_fract}			### only search if subvol fraction is greater than

### required parameters defining the basic the search process

disp_max		{disp_max}			### in voxels, used for range checking and global search limits
num_srch_dof		{num_srch_dof}			### 3, 6, or 12
obj_function		{obj_function}			### sad, ssd, zssd, nssd, znssd 
interp_type		{interp_type}		### trilinear, tricubic

### optional parameters tuning and refining the search process

rigid_trans		{rigid_trans}		### rigid body offset of target volume, in voxels
basin_radius		{basin_radius}			### coarse-search resolution, in voxels, 0.0 = none
subvol_aspect		{subvol_aspect}		### subvolume aspect ratio
num_points_to_process   {num_points_to_process}  ### Number of points in the point cloud to process
starting_point  {starting_point}    ### x,y,z location of starting point for DVC analysis
'''

def update_progress(main_window, process, total_points, required_runs, run_succeeded, start_time, num_points_to_process):
    # print("Required runs", required_runs)
    while(process.canReadLine()):
        #print("READ")
        string = process.readLine()  
        line = str(string, "utf-8")
        
        global count
        count+=1
        #print("{:.0f} \n".format(count/(total_points+3*required_runs)*100))
        if hasattr(main_window, 'progress_window'):
            #print(count/(total_points+3*required_runs)*100)
            # main_window.progress_window.setValue(count/(total_points+3*required_runs)*100)

            try:
                # try to infer the number of points processed from the output of the dvc executable
                num_processed_points = int(string.split('/')[0])
                etcs = 0
                prog = 0
                if num_processed_points > 0:
                    prog = int(num_processed_points/num_points_to_process*100)-1
                    etcs = ((time.time()-start_time) * num_points_to_process / num_processed_points) - (time.time()-start_time)
                try:
                    etc = time.strftime("%H:%M:%S s", time.gmtime(etcs))
                except:
                    etc = 'Error estimating time to completion'

                ETC_line = "\nEstimated time to completion of this step {} ".format(etc)
            except ValueError:
                num_processed_points = 0
                prog = 0
                ETC_line = ''
                
            main_window.progress_window.setValue(prog)
            label_text = main_window.progress_window.labelText().split("\n")[0]
            main_window.progress_window.setLabelText(
                    "{}\n{}{}".format(label_text, line, ETC_line))
                
        if line[:11] == "Input Error":
            run_succeeded = False
            if hasattr(main_window, 'progress_window'):
                main_window.progress_window.setValue(100)
            if hasattr(main_window, 'alert'):
                main_window.alert.close() #finish fn triggers before update_progress so has already made an alert saying it succeeded
            displayFileErrorDialog(main_window, line, "Error")
            process.kill()
            return
    #sys.stdout.flush()

def create_progress_window(main_window, title, text, max = 100, cancel = None):
        main_window.progress_window = QProgressDialog(text, "Cancel", 0,max, main_window, QtCore.Qt.Window) 
        main_window.progress_window.setWindowTitle(title)
        main_window.progress_window.setWindowModality(QtCore.Qt.ApplicationModal) #This means the other windows can't be used while this is open
        main_window.progress_window.setMinimumDuration(0.1)
        main_window.progress_window.setWindowFlag(QtCore.Qt.WindowCloseButtonHint, False)
        main_window.progress_window.setWindowFlag(QtCore.Qt.WindowMaximizeButtonHint, False)
        if cancel is None:
            main_window.progress_window.setCancelButton(None)
        else:
            main_window.progress_window.canceled.connect(cancel)

def displayFileErrorDialog(main_window, message, title):
    msg = QMessageBox(main_window)
    msg.setIcon(QMessageBox.Critical)
    msg.setWindowTitle(title)
    msg.setText(message)
    #msg.setDetailedText(main_window.e.ErrorMessage())
    msg.exec_()

def cancel_run(main_window, process, run_succeeded):
    
    if not run_succeeded:
        # print ("run cancelled?")
        process.kill()
        main_window.alert = QMessageBox(QMessageBox.NoIcon,"Cancelled","The run was cancelled.", QMessageBox.Ok)  
        main_window.alert.show()
        run_succeeded = False
    # else:
    #     print ("all OK, all processes ended")


def finished_run(main_window, exitCode, exitStatus, process = None, required_runs = 1, run_succeeded = False, finish_fn = None):
    global runs_completed
    runs_completed+=1
    # print ("DVC command ended.", exitCode, exitStatus)
    # print("Completed, ", runs_completed)

    if run_succeeded:
        if runs_completed == required_runs:
            if hasattr(main_window, 'progress_window'):
                main_window.progress_window.close()
            main_window.alert = QMessageBox(QMessageBox.NoIcon,"Success","The DVC code ran successfully.", QMessageBox.Ok) 
            main_window.alert.show()
            if finish_fn is not None:
                finish_fn()
            runs_completed = 0
            global count
            count = 0
    #print ("DVC command ended.", exitCode, exitStatus)
    

    # if( runs_completed >=  required_runs): #and  cancelled == False):
    #     runs_completed = 0
    # print("did")

class DVC_runner(object):
    def __init__(self, main_window, input_file, finish_fn, run_succeeded, session_folder):
        # print("The session folder is", session_folder)
        self.main_window = main_window
        self.input_file = input_file
        self.finish_fn = finish_fn
        self.run_succeeded = run_succeeded
        self.session_folder = session_folder

    def set_up(self, *args, **kwargs):
        '''This function sets up the DVC run, creating the run configurations and setting up the run folders.
        
        Parameters:
        -----------
        *args: not used
        **kwargs: 
            message_callback: callback function for messages
            progress_callback: callback function for progress updates

        Returns:
        --------
        None

        This function reads a json file containing the run configuration and creates the run configurations.
        The json file contains the following:
        - subvolume_points: list of number of points to process at each subvolume
        - subvolume_sizes: list of subvolume sizes, if not an integer multiple runs will be created
        - points: number of points to process in the point cloud
        - roi_files: list of point cloud files
        - reference_file: reference image volume
        - correlate_file: correlation image volume
        - vol_bit_depth: bit depth of the image volumes. If loading an image not of type int8 or int16, the image will be converted to uint16.
        - vol_hdr_lngth: header length of the image volumes
        - dims: dimensions of the image volumes
        - subvol_geom: geometry of the subvolume, either cube or sphere
        - subvol_npts: number of points in the subvolume, not used
        - disp_max: maximum displacement in voxels
        - dof: number of parameters in the optimisation, 3,6, or 12
        - obj: objective function, either sad, ssd, zssd, nssd, or znssd
        - interp_type: interpolation type, either trilinear or tricubic
        - run_folder: folder to save the run results and configurations
        - rigid_trans: rigid translation between the reference and correlation volumes
        - point0_world_coordinate: world coordinates of the starting point for the DVC analysis
        '''

        self.processes = []
        self.process_num = 0
        message_callback = kwargs.get('message_callback', PrintCallback())
        progress_callback = kwargs.get('progress_callback', PrintCallback())  
        
        # created in dvc_interface create_run_config
        with open(self.input_file) as tmp:
            config = json.load(tmp)

        subvolume_points = config['subvolume_points'] 
        subvolume_sizes = config['subvolume_sizes']
        points = int(config['points'])
        num_points_to_process = points

        roi_files = config['roi_files']
        reference_file = config['reference_file']
        correlate_file = config['correlate_file']
        
        progress_callback.emit(10)
        # Convert to raw if files are a list of tiffs
        if isinstance(reference_file, (list, tuple)):
            message_callback.emit("Converting reference file to raw format")
            base = os.path.abspath(self.session_folder)
            raw_reference_file_fname = os.path.join(base, config['run_folder'], 'reference.raw')
            save_tiff_stack_as_raw(reference_file, raw_reference_file_fname, progress_callback, 10, 50)
            reference_file = raw_reference_file_fname
        progress_callback.emit(50)

        if isinstance(correlate_file, (list, tuple)):
            message_callback.emit("Converting correlate file to raw format")
            base = os.path.abspath(self.session_folder)
            raw_correlate_file_fname = os.path.join(base, config['run_folder'], 'correlate.raw')
            save_tiff_stack_as_raw(correlate_file, raw_correlate_file_fname, progress_callback, 50, 90)
            correlate_file = raw_correlate_file_fname
        progress_callback.emit(90)

        message_callback.emit("Creating run configurations")
        vol_bit_depth = int(config['vol_bit_depth'])
        if vol_bit_depth not in [8, 16]:
            # the data will be converted to 16 bit by save_tiff_stack_as_raw
            # it won't work with other formats
            vol_bit_depth = 16
        vol_hdr_lngth = int(config['vol_hdr_lngth'])

        if 'vol_endian' in config:
            endian = config['vol_endian']
        else:
            endian = None

        dims= config['dims'] #image dimensions

        subvol_geom = config['subvol_geom']

        subvol_npts = config['subvol_npts']

        disp_max = int(config['disp_max'][0]) #TODO: need to change
        dof = int(config['dof'])
        obj = config['obj']
        interp_type = config['interp_type']

        rigid_trans = config['rigid_trans']
        starting_point = config['point0_world_coordinate']

        # Change directory into the folder where the run will be saved:
        os.chdir(self.session_folder)
        # this is the one directory we created where we will run the dvc command in
        # we want to change this to create multiple directories first and then run through
        # all the directory created https://github.com/TomographicImaging/iDVC/issues/37
        # see also https://github.com/TomographicImaging/iDVC/pull/69
        self.run_folder = config['run_folder']

        #running the code:

        # this should be renamed to num_optimisations
        required_runs = len(subvolume_points)*len(subvolume_sizes)   

        # check for the extension
        if platform.system() in ['Linux', 'Darwin']:    
            exe_file = 'dvc'
        elif platform.system() == 'Windows':
            exe_file = 'dvc.exe'
        else:
            raise ValueError('Not supported platform, ', platform.system())
        
        total_points = 0
        for cloud in roi_files:
            i=0
            with open(cloud) as f:
                for i, l in enumerate(f):
                    pass
            i+=1
            #print(i)
            if i < points:
                total_points += i
            else:
                total_points += points

        total_points = total_points * len(subvolume_points)

        #print("Total points", total_points)

        #process.readyRead.connect(lambda: update_progress(main_window, process, total_points, 
        #    required_runs, run_succeeded))
        

        start_progress = 90
        end_progress = 99
        file_count = -1
        # point0 = main_window.getPoint0WorldCoords()
            
        for roi_num, roi_file in enumerate(roi_files):
            file_count +=1
            subvolume_size = int(subvolume_sizes[file_count])
            #print(roi_file)
            distances = []

            for subv_num, subvolume_point in enumerate(subvolume_points):
                now = datetime.now()
                # use a counter for both for loops
                counter = subv_num + roi_num * len(subvolume_points)
                
                this_run_folder = os.path.join(self.run_folder, "dvc_result_{}".format(counter))
                os.mkdir(this_run_folder)
                output_filename = os.path.join(this_run_folder, "dvc_result_{}".format(counter))
                config_filename = os.path.join(this_run_folder,"dvc_config.txt")
                
                grid_roi_fname = os.path.join(this_run_folder, "grid_input.roi")
                # copies the pointcloud file as a whole in the run directory
                try:
                    shutil.copyfile(roi_file, grid_roi_fname)
                except Exception as err:
                    # this is not really a nice way to open an error message!
                    self.main_window.displayFileErrorDialog(message=str(err), title="Error creating config files")
                    return
                
                
                config =  blank_config.format(
                    reference_filename=  reference_file, # reference tomography image volume
                    correlate_filename=  correlate_file, # correlation tomography image volume
                    point_cloud_filename = grid_roi_fname,
                    output_filename= output_filename,
                    vol_bit_depth=  vol_bit_depth, # get from ref, 8 or 16
                    vol_hdr_lngth=vol_hdr_lngth,# get from ref, fixed-length header size, may be zero
                    vol_wide= dims[0], # number of x slices
                    vol_high= dims[1], # number of y slices
                    vol_tall= dims[2], #number of z slices
                    vol_endian = endian,
                    subvol_geom=  subvol_geom,
                    subvol_size=  subvolume_size, 
                    subvol_npts= subvolume_point,
                    subvol_thresh='off',
                    gray_thresh_min='27',
                    gray_thresh_max='127',
                    min_vol_fract='0.2',
                    disp_max=  disp_max, #38 for test image
                    num_srch_dof=  dof, #6 for test image
                    obj_function=  obj, 
                    interp_type=  interp_type, #tricubic for test image
                    rigid_trans= rigid_trans, #translation between ref and cor - determined from image registration
                    basin_radius='0.0',
                    subvol_aspect='1.0 1.0 1.0',# image spacing
                    num_points_to_process=num_points_to_process, 
                    starting_point='{} {} {}'.format(*starting_point)) 
                time.sleep(1)
                with open(config_filename,"w") as config_file:
                    config_file.write(config)
            
                #if run_count == len( subvolume_points):
                # process.start( exe_file , [ param_file ] )          
                cancelled=False

                # TODO: 
                # ideally we should start a new process from the main thread once the previous
                # has finished
                # wait for process to finish before doing next run
                
                # process.waitForFinished(msecs=2147483647)
                self.processes.append( 
                    (exe_file, [ config_filename ], required_runs, total_points, num_points_to_process)
                )
                progress_callback.emit(int(start_progress + (end_progress - start_progress) * (subv_num / len(roi_files))))
            progress_callback.emit(100)
                
        
    def run_dvc(self, **kwargs):
        main_window = self.main_window
        input_file = self.input_file
        finish_fn = self.finish_fn
        run_succeeded = self.run_succeeded
        
        process = QtCore.QProcess()
        
        env = QtCore.QProcessEnvironment.systemEnvironment()
        try:
            nthreads = main_window.settings.value('omp_threads')
        except Exception as err:
            nthreads = '4'
            print (err)
        env.insert("OMP_NUM_THREADS", nthreads)
        process.setProcessEnvironment(env)

        # print("Processes: ", self.processes)
        # print("num: ", self.process_num)        

        # start time
        start_time = time.time()

        exe_file, param_file, required_runs,\
            total_points, num_points_to_process = self.processes[self.process_num]

        process.setWorkingDirectory(os.getcwd())
        # process.finished.connect(partial(finished_run, main_window, 
        #             process = process, required_runs = required_runs, 
        #             run_succeeded = run_succeeded, finish_fn = finish_fn))
        process.finished.connect(self.finished_run)
        process.started.connect(self.onStarted)
        if self.process_num == 0:
            main_window.create_progress_window("Running", 
                "Running DVC code {}/{}".format(self.process_num +1, len(self.processes)), 100,
                #lambda: cancel_run(main_window, process, run_succeeded)
                lambda: self.onCancel(process)
                )
        else:
            main_window.progress_window.setLabelText(
                "Running DVC code {}/{}".format(self.process_num +1, len(self.processes))
            )
            main_window.progress_window.canceled.connect(lambda: self.onCancel(process))
        process.readyRead.connect(
            lambda: update_progress(main_window, process, total_points, required_runs,\
                                    run_succeeded, start_time, num_points_to_process))
        # process.finished.connect(self.run_dvc())
        process.start( exe_file , param_file )

    def onStarted(self):
        pass

    def onCancel(self, process):
        main_window = self.main_window
        run_succeeded = self.run_succeeded
        state = process.state()
        # print ("Process state", state)
        if state in [2,1]:
            # print ("Cancelling run")
            process.kill()
            main_window.alert = QMessageBox(QMessageBox.NoIcon,"Cancelled","The run was cancelled.", QMessageBox.Ok)  
            main_window.alert.show()
            self.run_succeeded = False
        elif state == 0:
            print ("all OK, all processes ended")

    def finished_run(self, exitCode, exitStatus):
        main_window = self.main_window
        input_file = self.input_file
        finish_fn = self.finish_fn
        run_succeeded = self.run_succeeded
        required_runs = self.processes[self.process_num][2]

        print("finished {}/{} (or {}) with {} {}"
              .format(self.process_num + 1, required_runs,
                len(self.processes), exitCode, exitStatus))
        
        if exitStatus == 0:
            self.run_succeeded = self.run_succeeded and True
        else:
            self.run_succeeded = self.run_succeeded and False
            
        if self.process_num == required_runs-1:
            main_window.progress_window.close()
            if self.run_succeeded:
                main_window.alert = QMessageBox(QMessageBox.NoIcon,
                    "Success","The DVC code ran successfully.", QMessageBox.Ok)
            else:
                main_window.alert = QMessageBox(QMessageBox.NoIcon,
                    "Fail","The DVC code had some troubles.", QMessageBox.Ok) 
            main_window.alert.show()
            if finish_fn is not None:
                finish_fn() 
        else:
            self.process_num += 1
            self.run_dvc()

            


import PySide2
from PySide2 import QtWidgets, QtGui
import os, sys
import logging
import argparse


def main():

    parser = argparse.ArgumentParser(description='iDVC - Digital Volume Correlation Software')

    parser.add_argument('--debug', type=str)
    args = parser.parse_args()

    if args.debug in ['debug', 'info', 'warning', 'error', 'critical']:
        level = eval(f'logging.{args.debug.upper()}')
        logging.basicConfig(level=level)
        logging.info(f"iDVC: Setting debugging level to {args.debug.upper()}")
    
    app = QtWidgets.QApplication([])
    
    file_dir = os.path.dirname(__file__)
    owl_file = os.path.join(file_dir, "DVCIconSquare.png")
    owl = QtGui.QPixmap(owl_file)
    splash = QtWidgets.QSplashScreen(owl)
    splash.show()
    

    import vtk
    err = vtk.vtkFileOutputWindow()
    err.SetFileName("../viewer.log")
    vtk.vtkOutputWindow.SetInstance(err)

    from idvc.dvc_interface import MainWindow
    window = MainWindow()
    
    window.show()
    splash.finish(window)

    

    sys.exit(app.exec_())

if __name__ == "__main__":
    main()



import glob
import os

import ccpi.viewer.viewerLinker as vlink

import vtk
from ccpi.viewer import viewer2D, viewer3D
from ccpi.viewer.CILViewer2D import (SLICE_ORIENTATION_XY,
                                     SLICE_ORIENTATION_XZ,
                                     SLICE_ORIENTATION_YZ)
from ccpi.viewer.QCILViewerWidget import QCILViewerWidget
from ccpi.viewer.utils import *
from ccpi.viewer.utils.error_handling import ErrorObserver
from PySide2 import QtCore, QtGui, QtWidgets
from PySide2.QtCore import *
from PySide2.QtGui import *
from PySide2.QtWidgets import *

from idvc.ui.widgets import *

from idvc.utilities import RunResults
import logging 


class VisualisationWindow(QtWidgets.QMainWindow):
    '''creates a window which will contain the VisualisationWidgets
    '''
    def __init__(self, parent):
        super().__init__()
        self.parent = parent
        self.setMinimumSize(200,200)

    def createPopupMenu(self):
        '''returns an empty menu for the main window to use as a popup menu.
        
        https://doc.qt.io/qt-6/qmainwindow.html#createPopupMenu
        '''
        return QtWidgets.QMenu(self)
    
class VisualisationWidget(QtWidgets.QMainWindow):
    '''creates a window with a QCILViewerWidget as the central widget
    '''
    def __init__(self, parent, viewer=viewer2D, interactorStyle=vlink.Linked2DInteractorStyle, enableSliderWidget=True):
        super().__init__()
        self.parent = parent

        self.e = ErrorObserver()
        self.viewer = viewer
        self.interactorStyle = interactorStyle
        self.enableSliderWidget = enableSliderWidget
        self.createEmptyFrame()
        self.threadpool = QThreadPool()
        self._consume_CharEvent = ['s', 'w']
        

    def getViewer(self):
        return self.frame.viewer

    def getInteractor(self):
        return self.getViewer().getInteractor()

    def getInteractorStyle(self):
        return self.getViewer().style

    def getViewerType(self):
        return self.viewer
     
    def createEmptyFrame(self):
        #print("empty")
        self.frame = QCILViewerWidget(self.parent, self.viewer, shape=(600,600), interactorStyle=self.interactorStyle, 
                                      enableSliderWidget=self.enableSliderWidget)
        self.setCentralWidget(self.frame)
        self.image_file = [""]
       
    def displayImageData(self):
        self.createEmptyFrame()
        if self.viewer == viewer3D:
            #set volume mapper according to user settings:
            if self.parent.settings.value("volume_mapper") == "cpu":
                self.frame.viewer.volume_mapper = vtk.vtkFixedPointVolumeRayCastMapper()
                self.frame.viewer.volume.SetMapper(self.frame.viewer.volume_mapper)
        else:
            self.frame.viewer.setVisualisationDownsampling(self.parent.resample_rate)
            self.frame.viewer.setDisplayUnsampledCoordinates(True)

            vs_widgets = self.parent.visualisation_setting_widgets

            vs_widgets['loaded_image_dims_value'].setVisible(True)
            vs_widgets['loaded_image_dims_value'].setText(str(self.parent.unsampled_image_dimensions))

            #print("resample rate: ", self.parent.resample_rate)

            if self.parent.resample_rate != [1,1,1]:
                vs_widgets['displayed_image_dims_value'].setVisible(True)
                vs_widgets['displayed_image_dims_label'].setVisible(True)
                #print("Disp image size ", [self.parent.ref_image_data.GetDimensions()[i] for i in range(3)])
                vs_widgets['displayed_image_dims_value'].setText(str([round(self.parent.ref_image_data.GetDimensions()[i]) for i in range(3)]))
                vs_widgets['coords_combobox'].setEnabled(True)
                vs_widgets['coords_combobox'].setCurrentIndex(0)
                vs_widgets['coords_warning_label'].setVisible(True)
                vs_widgets['coords_info_label'].setVisible(True)

            
            else:
                vs_widgets['displayed_image_dims_value'].setVisible(False)
                vs_widgets['displayed_image_dims_label'].setVisible(False)
                vs_widgets['coords_warning_label'].setVisible(False)
                vs_widgets['coords_info_label'].setVisible(False)

                vs_widgets['coords_combobox'].setEnabled(False)
                vs_widgets['coords_combobox'].setCurrentIndex(0)

        # consume 's' and 'w' events
        # https://github.com/vais-ral/CILViewer/issues/332#issuecomment-1888940327
        self.getInteractor().AddObserver('CharEvent', self.consumeCharEvent, 10.)

        self.frame.viewer.setInput3DData(self.image_data)  
        interactor = self.frame.viewer.getInteractor()


        if hasattr(self.parent, 'orientation'):
                orientation = self.parent.orientation
        else:
            orientation = self.frame.viewer.getSliceOrientation()
        
        if orientation == SLICE_ORIENTATION_XZ:
            axis = 'y'
        elif orientation == SLICE_ORIENTATION_YZ:
            axis = 'x'
        else:
            axis = 'z'
        interactor.SetKeyCode(axis)

        if self.viewer == viewer2D:
            self.frame.viewer.style.OnKeyPress(interactor, 'KeyPressEvent')
            if self.parent.current_slice:
                if self.parent.current_slice <= self.frame.viewer.img3D.GetExtent()[self.frame.viewer.getSliceOrientation()*2+1]:
                    self.frame.viewer.displaySlice(self.parent.current_slice)


        if self.viewer == viewer3D:
            self.frame.viewer.style.OnKeyPress(interactor, 'KeyPressEvent')
            # Depth peeling for volumes doesn't work as we would like when we have the vtk.vtkFixedPointVolumeRayCastMapper() instead of the vtk.vtkSmartVolumeMapper()
            # self.frame.viewer.sliceActor.GetProperty().SetOpacity(0.99)
            # self.frame.viewer.ren.SetUseDepthPeeling(True)
            # self.frame.viewer.renWin.SetAlphaBitPlanes(True)
            # self.frame.viewer.renWin.SetMultiSamples(False)
            # self.frame.viewer.ren.UseDepthPeelingForVolumesOn()
    
            if self.parent.current_slice:
                if self.parent.current_slice <= self.frame.viewer.img3D.GetExtent()[self.frame.viewer.getSliceOrientation()*2+1]:
                    self.frame.viewer.style.SetActiveSlice(self.parent.current_slice)
                    self.frame.viewer.style.UpdatePipeline()

        # print("set input data for" + str(self.viewer))

        if self.viewer == viewer2D:
            self.PlaneClipper = cilPlaneClipper()
            self.PlaneClipper.SetInteractorStyle(self.frame.viewer.style)


    def setImageData(self, image_data):
        self.image_data = image_data

    def getImageData(self):
        return self.image_data

    def consumeCharEvent(self, interactor, event):
        '''Changes the event in self._consume_CharEvent to "" so that the event is not processed by the viewer.
        
        This prevents the viewer from changing the rendered scene to surface or wireframe mode when the user presses 's' or 'w' respectively.
        '''
        if interactor.GetKeyCode() in self._consume_CharEvent:
            logging.info("Consuming event: " + interactor.GetKeyCode())
            interactor.SetKeyCode("")

class GraphsWindow(QMainWindow):
    '''creates a new window with graphs from results saved in the selected run folder.
    '''
    def __init__(self, parent=None):
        super(GraphsWindow, self).__init__(parent)
        self.setWindowTitle("Digital Volume Correlation Results")
        DVCIcon = QtGui.QIcon()
        DVCIcon.addFile("DVCIconSquare.png")

        # Menu
        self.menu = self.menuBar()
        self.file_menu = self.menu.addMenu("File")
        self.settings_menu = self.menu.addMenu("Settings")

        displacement_setting_action = QAction("Show Displacement Relative to Reference Point 0", self)
        displacement_setting_action.setCheckable(True)
        displacement_setting_action.setChecked(False)
        self.displacement_setting_action = displacement_setting_action

        displacement_setting_action.triggered.connect(self.ReloadGraphs)
        self.settings_menu.addAction(displacement_setting_action)


        # Exit QAction
        exit_action = QAction("Exit", self)
        exit_action.setShortcut(QKeySequence.Quit)
        exit_action.triggered.connect(self.close)
        self.file_menu.addAction(exit_action)

        #Tab positions:
        self.setTabPosition(QtCore.Qt.AllDockWidgetAreas,QTabWidget.North)
        self.setDockOptions(QMainWindow.ForceTabbedDocks)
             
        # Window dimensions
        geometry = qApp.desktop().availableGeometry(self)

        self.setGeometry(50,50, geometry.width()-100, geometry.height()-100)
        #self.setFixedSize(geometry.width() * 0.6, geometry.height() * 0.8)

    def SetResultsFolder(self, folder):
        self.results_folder = folder
        self.setWindowTitle("Digital Volume Correlation Results - {foldername}".format(foldername=os.path.basename(self.results_folder)))
    
    def ReloadGraphs(self):
        self.DeleteAllWidgets()
        self.CreateDockWidgets(displ_wrt_point0 = self.displacement_setting_action.isChecked())

    def DeleteAllWidgets(self):
         for current_dock in self.findChildren(QDockWidget):
            current_dock.close()
            del current_dock

    def CreateDockWidgets(self, displ_wrt_point0 = False):
        result_list=[]
        #print(results_folder[0])
        for folder in glob.glob(os.path.join(self.results_folder, "dvc_result_*")):
            file_path = os.path.join(folder, os.path.basename(folder))
            result = RunResults(file_path)
            result_list.append(result)
    
            GraphWidget = SingleRunResultsWidget(self, result, displ_wrt_point0)
            dock1 = QDockWidget(result.title,self)
            dock1.setFeatures(QDockWidget.NoDockWidgetFeatures) 
            dock1.setAllowedAreas(QtCore.Qt.RightDockWidgetArea)
            dock1.setWidget(GraphWidget)
            self.addDockWidget(QtCore.Qt.RightDockWidgetArea,dock1)
    
        prev = None

        for current_dock in self.findChildren(QDockWidget):
            if self.dockWidgetArea(current_dock) == QtCore.Qt.RightDockWidgetArea:
                existing_widget = current_dock

                if prev:
                    self.tabifyDockWidget(prev,current_dock)
                prev= current_dock
        
        SummaryTab = SummaryGraphsWidget(self, result_list)
        dock = QDockWidget("Summary",self)
        dock.setFeatures(QDockWidget.NoDockWidgetFeatures) 
        dock.setAllowedAreas(QtCore.Qt.RightDockWidgetArea)
        dock.setWidget(SummaryTab)
        self.addDockWidget(QtCore.Qt.RightDockWidgetArea,dock)
        self.tabifyDockWidget(prev,dock)

        dock.raise_() # makes summary panel the one that is open by default.

        # Stop the widgets in the tab to be moved around
        for wdg in self.findChildren(QTabBar):
            wdg.setMovable(False)




import logging
import os

import matplotlib.pyplot as plt
import numpy as np
import scipy.signal
from matplotlib.patches import Rectangle


class AutomaticRegistration:

    def __init__(self,
                 im0,
                 im1,
                 p3d_0,
                 size,
                 log_folder=None,
                 err_thresh=2,
                 max_iterations=30):
        """
        Initialize an object for automatic registration of images.

        Parameters
        ----------
        im0 : numpy.ndarray
            The reference image (3D numpy array).
        im1 : numpy.ndarray
            The image to be registered (3D numpy array).
        p3d_0 : list[int, int, int]
            The starting point for the registration process.
        size : numpy.ndarray
            The size of the volume in which the registration is performed.
            Format: [[int, int], [int, int], [int, int]]
        log_folder : str, optional
            The folder path where the log file for automatic registration will be saved.
            If not provided, no logging will be saved.
        err_thresh : int, optional
            The error threshold allowed by the registration procedure in each direction (in pixels).
        max_iterations : int, optional
            The maximum number of iterations allowed for the registration process.
        """
        # image0 and image1 are redefined in `run`
        self.image0 = im0
        self.image1 = im1
        self.p3d_0 = p3d_0
        self.size = size
        self.edge = self.calc_edge()
        datatype = im0.dtype
        if datatype == 'int8' or datatype == 'uint8' or datatype == '>u1' or datatype == '<u1':
            self.datatype_slice = 'int16'
        elif datatype == 'int16' or datatype == 'uint16' or datatype == '>u2' or datatype == '<u2':
            self.datatype_slice = 'int32'
        self.err_thresh = err_thresh
        self.max_iterations = max_iterations
        if log_folder is not None:
            self.log_folder = log_folder
            logger = logging.getLogger()
            file_handler = logging.FileHandler(
                filename=log_folder + r'\automatic_registration_logging.log',
                mode='w')
            formatter = logging.Formatter("%(asctime)s %(message)s")
            file_handler.setFormatter(formatter)
            logger.setLevel(logging.INFO)
            logger.addHandler(file_handler)

    def calc_edge(self):
        """
        Given the vertices position of the parallelepiped, it outputs an array with the edges
        values (in pixels) plus the first and second edge repeated.
        The repetition is to simplify the coding in other methods.

        Returns:
            list: A list containing the edge lengths of the parallelepiped.
        """
        edge = []
        for dim in range(0, 3):
            edge.append(abs(self.size[dim, 0] - self.size[dim, 1]))
        edge.append(edge[0])
        edge.append(edge[1])
        return edge

    def cross_image(self, sel0, sel1):
        """
        Calculate cross image of two 2D images, they can have any size as
        long as they are equal size.

        Parameters
        ------------
        sel0 : 2D numpy.array
        sel1 : 2D numpy.array

        Returns
        --------
        corr_img : 2D numpy.array
            The correlation matrix of `sel0`, `sel1`.
        """
        sel0_float = sel0.astype('float')
        sel1_float = sel1.astype('float')
        sel0_float -= np.mean(sel0_float)
        sel1_float -= np.mean(sel1_float)
        # calculate the correlation image by doing a convolution and flipping of one of the images,
        # this is faster than correlate2D
        # ::-1 returns a new list or string with all elements in reverse order
        corr_img = scipy.signal.fftconvolve(sel0_float,
                                            sel1_float[::-1, ::-1],
                                            mode='same')
        return corr_img

    def func_slicing_uservolume(self, im0, im1, selected_slice, dimension):
        """
        Function to output the slices at position `selectedslice` of two 3D images.
        Selects the slice in a smaller volume defined by the user with `self.size`.
        This function also changes the type of the data so the values can be signed
        when calculating the difference diff.

        Parameters
        ------------
        im0 : 3D numpy.array
        im1 : 3D numpy.array
        selected_slice : int
            The position of the slice along the `dimension` axis in the 3D arrays.
        dimension : int {`0`, `1`, `2`}
            Represents one of the xyz spatial axis in the 3D images.

        Returns
        --------
        sel0 : 2D numpy.array
            The selected slice in `im0`.
        sel1 : 2D numpy.array
            The selected slice in `im1`.
        diff : 2D numpy.array
            The difference between `sel0` and `sel1`.
        """
        size = self.size
        datatype_slice = self.datatype_slice

        if dimension == 0:
            sel0 = im0[selected_slice, size[1, 0]:size[1, 1],
                       size[2, 0]:size[2, 1]].astype(datatype_slice)
            sel1 = im1[selected_slice, size[1, 0]:size[1, 1],
                       size[2, 0]:size[2, 1]].astype(datatype_slice)
        elif dimension == 1:
            sel0 = im0[size[0, 0]:size[0, 1], selected_slice,
                       size[2, 0]:size[2, 1]].astype(datatype_slice)
            sel1 = im1[size[0, 0]:size[0, 1], selected_slice,
                       size[2, 0]:size[2, 1]].astype(datatype_slice)
        elif dimension == 2:
            sel0 = im0[size[0, 0]:size[0, 1], size[1, 0]:size[1, 1],
                       selected_slice].astype(datatype_slice)
            sel1 = im1[size[0, 0]:size[0, 1], size[1, 0]:size[1, 1],
                       selected_slice].astype(datatype_slice)

        diff = sel0 - sel1

        return sel0, sel1, diff

    def func_slicing(self, im0, im1, selected_slice, dimension):
        """
        Function to output the slices at position `selected_slice` of two 3D images.
        This function also changes the type of the data so the values can be signed
        when calculating the difference diff.

        Parameters
        ----------
        im0 : numpy.ndarray
            The first 3D image.
        im1 : numpy.ndarray
            The second 3D image.
        selected_slice : int
            The position of the slice along the specified `dimension` in the 3D arrays.
        dimension : int {0, 1, 2}
            The axis representing one of the xyz spatial dimensions in the 3D images.

        Returns
        -------
        sel0 : numpy.ndarray
            The selected slice in `im0`.
        sel1 : numpy.ndarray
            The selected slice in `im1`.
        diff : numpy.ndarray
            The difference between `sel0` and `sel1`.
        """
        datatype_slice = self.datatype_slice

        if dimension == 0:
            sel0 = im0[selected_slice, :, :].astype(datatype_slice)
            sel1 = im1[selected_slice, :, :].astype(datatype_slice)
        elif dimension == 1:
            sel0 = im0[:, selected_slice, :].astype(datatype_slice)
            sel1 = im1[:, selected_slice, :].astype(datatype_slice)
        elif dimension == 2:
            sel0 = im0[:, :, selected_slice].astype(datatype_slice)
            sel1 = im1[:, :, selected_slice].astype(datatype_slice)

        diff = sel0 - sel1

        return sel0, sel1, diff

    def func_2D_displacement(self, sel0, sel1):
        """
        Calculates the maximum of the correlation matrix between two images and its position,
        which could represent the displacement of two shifted images.

        Parameters
        ----------
        sel0 : numpy.ndarray
            The first 2D image array.
        sel1 : numpy.ndarray
            The second 2D image array.

        Returns
        -------
        shift2D : list[int, int]
            The displacement of the two images with respect to the center of `sel0` and `sel1`.
            The convention for the sign of the displacement is considered.
        max : float
            The maximum value of the correlation matrix corresponding to the displacement.
        """
        corr_img = self.cross_image(sel0, sel1)
        displacement = np.array(
            np.unravel_index(np.argmax(corr_img), corr_img.shape))
        max = corr_img[displacement[0], displacement[1]]
        centre = np.array([round(sel0.shape[0] / 2), round(sel0.shape[1] / 2)])
        shift2D = centre - displacement
        return shift2D, max

    def shift_3D_arrays(self, im0, im1, shift3D):
        """
        Implements the shift of two 3D images by the quantity `shift3D`.
        The result could crop both images and change the size of the returned 3D arrays.

        Parameters
        ----------
        im0 : numpy.ndarray
            The first 3D image array.
        im1 : numpy.ndarray
            The second 3D image array.
        shift3D : numpy.ndarray or list[int, int, int]
            The amount of shift to be applied to the images along each axis.
            The sign convention is taken into account to shift the 3D
            arrays in the right directions.

        Returns
        -------
        tuple[numpy.ndarray, numpy.ndarray]
            A tuple containing the shifted 3D image arrays.
            The first element is the shifted version of `im0`.
            The second element is the shifted version of `im1`.
            The size of the arrays may be reduced depending on the shift values.
        """
        if shift3D[0] > 0:
            im0 = im0[:-shift3D[0], :, :]
            im1 = im1[shift3D[0]:, :, :]
        elif shift3D[0] < 0:
            im0 = im0[abs(shift3D[0]):, :, :]
            im1 = im1[:-abs(shift3D[0]), :, :]
        if shift3D[1] > 0:
            im0 = im0[:, :-shift3D[1], :]
            im1 = im1[:, shift3D[1]:, :]
        elif shift3D[1] < 0:
            im0 = im0[:, abs(shift3D[1]):, :]
            im1 = im1[:, :-abs(shift3D[1]), :]
        if shift3D[2] > 0:
            im0 = im0[:, :, :-shift3D[2]]
            im1 = im1[:, :, shift3D[2]:]
        elif shift3D[2] < 0:
            im0 = im0[:, :, abs(shift3D[2]):]
            im1 = im1[:, :, :-abs(shift3D[2])]
        return im0, im1

    def shift_point_zero(self, p3d, shift3D, im0):
        """
        Revaluate the coordinates of a fixed point zero, `p3d`, whose 3d images have
        been shifted by `shift_3D`. If the shift is negative, the value is reassigned
        by subtracting the shift. If the shift is positive, the value is reassigned to that
        corresponding to the maximum index for `im0` in the axis of the shift.

        Parameters
        ----------
        p3d : numpy.ndarray or list[int, int, int]
            The coordinates of the fixed point zero.
        shift3d : numpy.ndarray or list[int, int, int]
            The shift values in each dimension.
        im0 : numpy.ndarray
            The 3D numpy array representing the image.

        Returns
        -------
        numpy.ndarray or list[int, int, int]
            The updated coordinates of the fixed point zero.
        """
        for dim in range(0, 3):
            if shift3D[dim] < 0:
                p3d[dim] = p3d[dim] + shift3D[dim]
            if shift3D[dim] > 0 and p3d[dim] >= np.shape(im0)[dim]:
                p3d[dim] = np.shape(im0)[dim] - 1
        return p3d

    def iterative_func(self, im0, im1):
        """
        Iterates the shift calculation process between two 3D image arrays.

        This method isolates the steps that will be iterated in a while loop. It can be redefined to
        add additional functionality to the iterations, such as plotting.

        Parameters
        ----------
        im0 : numpy.ndarray
            The first 3D image array.
        im1 : numpy.ndarray
            The second 3D image array.

        Returns
        -------
        DD3d : numpy.ndarray
            The calculated 3D shift between im0 and im1, in units of pixels.
        err : numpy.ndarray
            The error on the shift calculation, in units of pixels.
        """
        DD3d, err = self.calc_shift(im0, im1)
        return DD3d, err

    def calc_shift(self, im0, im1):
        """
        Calculates the 3D shift between two 3D numpy arrays

        Parameters
        ----------
        im0 : numpy.ndarray
            The reference image.
        im1 : numpy.ndarray
            The target image.

        Returns
        -------
        DD3d : numpy.ndarray
            The 3D shift evaluated from im0 and im1, in units of pixels.
        err : numpy.ndarray
            The error on shift calculation in units of pixels.
        """

        logging.info("Size of ref is " + str(np.shape(im0)) + ".")
        logging.info("Size of image 1 is " + str(np.shape(im1)) + ".\n")

        # centre3d = np.array([
        #     round(im0.shape[0] / 2),
        #     round(im0.shape[1] / 2),
        #     round(im0.shape[2] / 2)
        # ])
        # initialise the array to store 2 displacements in 3 directions
        DD6d = []
        # store max values of correlation matrices in the 3 directions
        for dim in range(0, 3):
            [sel0,
             sel1] = self.func_slicing_uservolume(im0, im1, self.p3d_0[dim],
                                                  dim)[0:2]
            DD = self.func_2D_displacement(sel0, sel1)[0]
            DD6d.append(DD)

        DD6d = np.array(DD6d)
        logging.info(
            "Array of displacements from the correlation images in 3 directions is \n"
            + str(DD6d) + ".\n")
        logging.info("Couplets are " + str([DD6d[2, 0], DD6d[1, 0]]) +
                     str([DD6d[0, 0], DD6d[2, 1]]) +
                     str([DD6d[0, 1], DD6d[1, 1]]) + ".")
        DD3d = np.array([0, 0, 0])
        err = [
            abs(DD6d[2, 0] - DD6d[1, 0]),
            abs(DD6d[0, 0] - DD6d[2, 1]),
            abs(DD6d[0, 1] - DD6d[1, 1])
        ]
        if err[0] <= self.err_thresh:
            DD3d[0] = DD6d[2, 0]
        if err[1] <= self.err_thresh:
            DD3d[1] = DD6d[0, 0]
        if err[2] <= self.err_thresh:
            DD3d[2] = DD6d[1, 1]

        if sum(abs(DD3d)) == 0:
            DD3d = self.calc_alternative_shift(DD3d, DD6d)
        logging.info("Calculated shift for this cycle is " + str(DD3d) + ".\n")

        return DD3d, err

    def calc_alternative_shift(self, DD3d, DD6d):
        """
        Calculates the alternative shift in one direction. This method selects the
        minimum shift among the couplets whose elements have the same sign.

        Parameters:
        ----------
        DD3d : numpy.ndarray [int, int, int]
            The current shift in the 3 directions.
        DD6d : numpy.ndarray [int, int, int, int, int, int]
            The couplets of shifts in the 3 directions calculated as the position
            of the maximum in the correlation matrices.

        Returns:
        -------
        numpy.ndarray [int, int, int]
            The updated shift in the 3 directions.
                Only one of the values could be updated by this method.
        """
        tmp = np.array([
            self.choose_shift(DD6d[2, 0], DD6d[1, 0]),
            self.choose_shift(DD6d[0, 0], DD6d[2, 1]),
            self.choose_shift(DD6d[0, 1], DD6d[1, 1])
        ])
        uDD3d = abs(tmp)
        if sum(abs(uDD3d)) != 0:
            minval = np.min(uDD3d[np.nonzero(uDD3d)])
            index = np.where(uDD3d == minval)[0][0]
            DD3d[index] = tmp[index]
        return np.array(DD3d)

    def choose_shift(self, a, b):
        """
        Choose the shift between two numbers.

        The shift between two numbers is determined based on the following rules:
        - If the signs of the two numbers are the same, the shift is set to the
          minimum of their absolute values.
        - If the signs are different, the shift is set to 0.

        Parameters:
        -----------
        a : np.int
            The first number.
        b : np.int
            The second number.

        Returns:
        --------
        np.int
            The chosen shift value.
        """
        return np.sign(a) * np.min(abs(np.array([a, b]))) if np.sign(
            a) == np.sign(b) else 0

    def run(self):
        """
        Run the automatic registration algorithm.

        This method performs the automatic registration algorithm by iteratively
        shifting 3D arrays and updating the point zero. It accumulates the overall
        3D shift and calculates the relative error at each iteration. The algorithm
        stops when the accumulated shift is zero or the maximum number of iterations
        is reached.

        Returns:
        --------
            None
        """
        self.DD3d_accumulate = [0, 0, 0]
        for counter in range(0, self.max_iterations):
            logging.info("\n\nCycle number " + str(counter) + ".\n")
            DD3d, err = self.iterative_func(self.image0, self.image1)
            err_rel = (err[0] / self.image0.shape[0] +
                       err[1] / self.image0.shape[1] +
                       err[2] / self.image0.shape[2]) / 3
            self.DD3d_accumulate = np.add(self.DD3d_accumulate, DD3d)
            logging.info("Relative error is " + str(err_rel) + ".\n")
            self.image0, self.image1 = self.shift_3D_arrays(
                self.image0, self.image1, DD3d)
            self.p3d_0 = self.shift_point_zero(self.p3d_0, DD3d, self.image0)
            logging.info("New point zero is" + str(self.p3d_0) + ".\n")
            if sum(abs(DD3d)) == 0:
                break

        logging.info("\n\nResults:\n")
        DD3d, err = self.iterative_func(self.image0, self.image1)
        logging.info("Relative error is " + str(err_rel) + ".\n\n\n\n\n")

        logging.info("Overall shift is" + str(self.DD3d_accumulate) + ".\n")
        logging.info(
            '-----------------Automatic registration ends successfully.-----------------\n\n\n\n\n'
        )


class AutomaticRegistrationWithPlotting(AutomaticRegistration):

    def __init__(self, intermediate_plot, filename0, filename1, outputfolder,
                 save, *args, **kwargs):
        """
        Initializes an instance of the AutomaticRegistrationWithPlotting class.

        Parameters
        ----------
        intermediate_plot : bool
            Flag indicating whether to generate intermediate plots.
        filename0 : str
            Path to the first input file.
        filename1 : str
            Path to the second input file.
        outputfolder : str
            Path to the output folder.
        save : bool
            Flag indicating whether to save the output.
        *args :
            Variable length argument list.
        **kwargs :
            Arbitrary keyword arguments.
        """
        self.intermediate_plot = intermediate_plot
        self.filenames = [filename0, filename1]
        self.outputfolder = outputfolder
        self.save = save
        super(AutomaticRegistrationWithPlotting,
              self).__init__(*args, **kwargs)

    def iterative_func(self, im0, im1):
        """
        Redefines the iterative function in the parent class.
        As in the parent class, this method iterates the shift calculation process
        between two 3D image arrays.
        It adds additional functionality by plotting the intermediate results.

        Parameters
        ----------
        im0 : numpy.ndarray
            The first 3D image array.
        im1 : numpy.ndarray
            The second 3D image array.

        Returns
        -------
        DD3d : numpy.ndarray
            The calculated 3D shift between im0 and im1, in units of pixels.
        err : numpy.ndarray
            The error on the shift calculation, in units of pixels.
        """
        DD3d, err = self.calc_shift(im0, im1)
        self.plot_three_directions(im0, im1)
        return DD3d, err

    def intensity_plot_array(self, a0, a1, a2, a3, dim, displacementpar,
                             rect00, rect10):
        """
        This method plots four images: a0, a1, a2, and a3, along with their corresponding colorbars.
        It also adds rectangle patches to the first and second images.
        The plot is saved as a figure if the 'save' parameter is set to True.

        Parameters
        ----------
        a0 : numpy.ndarray
            The first image array.
        a1 : numpy.ndarray
            The second image array.
        a2 : numpy.ndarray
            The cross-correlation image array.
        a3 : numpy.ndarray
            The difference image array.
        dim : int
            The dimension along which the slice is taken.
        displacementpar : str
            The displacement parameter.
        rect00 : matplotlib.patches.Rectangle
            The rectangle patch for the first image.
        rect10 : matplotlib.patches.Rectangle
            The rectangle patch for the second image.

        Returns
        -------
        None
        """
        # needs transpose because dim 1 has inverted axes
        if dim == 1:
            a0 = np.transpose(a0)
            a1 = np.transpose(a1)
            a2 = np.transpose(a2)
            a3 = np.transpose(a3)
        self.plt = plt
        fig, axs = plt.subplots(2, 2, sharex=False, sharey=False)
        fig.suptitle("Results after displacement of " + displacementpar +
                     ". Slice along dimension " + str(dim))
        txt = "\nImage0 = " + self.filenames[
            0] + ".\nImage1 = " + self.filenames[1] + ".\npoint zero = " + str(
                self.p3d_0) + ". uservolume = " + str(self.size[0]) + str(
                    self.size[1]) + str(self.size[2])
        plt.figtext(0.5,
                    0.01,
                    txt,
                    wrap=True,
                    horizontalalignment='center',
                    fontsize=8)
        fig.tight_layout(h_pad=2)
        ax00 = axs[0, 0].imshow(a0,
                                origin='lower',
                                cmap='gray',
                                vmin=np.min(a0),
                                vmax=np.max(a0),
                                interpolation='none')
        axs[0, 0].set_title('Image 0')
        fig.colorbar(mappable=ax00, location='right', shrink=0.6)

        ax10 = axs[1, 0].imshow(a1,
                                origin='lower',
                                cmap='gray',
                                vmin=np.min(a1),
                                interpolation='none')
        axs[1, 0].set_title('Image 1')
        fig.colorbar(mappable=ax10, location='right', shrink=0.6)
        ax01 = axs[0, 1].imshow(a2,
                                origin='lower',
                                cmap='gray',
                                vmin=np.min(a2),
                                interpolation='none')
        axs[0, 1].set_title('Cross correlation')
        fig.colorbar(mappable=ax01, location='right', shrink=0.6)
        ax11 = axs[1, 1].imshow(a3,
                                origin='lower',
                                cmap='gray',
                                vmin=np.min(a3),
                                interpolation='none')
        axs[1, 1].set_title('Diff')
        fig.colorbar(mappable=ax11, location='right', shrink=0.6)
        axs[0, 0].set_ylabel('Dimension ' + str(np.mod(dim + 1, 3)))
        axs[1, 0].set_ylabel('Dimension ' + str(np.mod(dim + 1, 3)))
        axs[1, 0].set_xlabel('Dimension ' + str(np.mod(dim + 2, 3)))
        axs[1, 1].set_xlabel('Dimension ' + str(np.mod(dim + 2, 3)))
        # resize the figure to match the aspect ratio of the axes
        fig.set_size_inches(10, 12, forward=True)
        axs[0, 0].add_patch(rect00)
        axs[1, 0].add_patch(rect10)
        if self.save is True:
            fig.savefig(os.path.join(
                self.outputfolder,
                'full_figure_' + displacementpar + str(dim) + '.png'),
                        dpi=600)
        return

    def plot_three_directions(self, im0, im1):
        """
        Plots image slices indicated by `self.p3d_0` and `self.size` in the three dimensions.

        Parameters
        ----------
        im0 : numpy.ndarray
            The first 3D image array.
        im1 : numpy.ndarray
            The second 3D image array.

        Returns
        ------
        None
        """
        for dim in range(0, 3):
            [sel0, sel1,
             diff] = self.func_slicing_uservolume(im0, im1, self.p3d_0[dim],
                                                  dim)
            corr_img = self.cross_image(sel0, sel1)
            if self.intermediate_plot is True:
                # stored to plot the large images too
                [largesel0,
                 largesel1] = self.func_slicing(im0, im1, self.p3d_0[dim],
                                                dim)[0:2]
                # rect must be repeated otherwise it gives errors when plotting
                rect1 = Rectangle(
                    (self.size[np.mod(dim + 2, 3),
                               0], self.size[np.mod(dim + 1, 3), 0]),
                    self.edge[dim + 2],
                    self.edge[dim + 1],
                    linewidth=1,
                    edgecolor='r',
                    facecolor="none")
                rect2 = Rectangle(
                    (self.size[np.mod(dim + 2, 3),
                               0], self.size[np.mod(dim + 1, 3), 0]),
                    self.edge[dim + 2],
                    self.edge[dim + 1],
                    linewidth=1,
                    edgecolor='r',
                    facecolor="none")
                self.intensity_plot_array(largesel0, largesel1,
                                          corr_img, diff, dim,
                                          str(self.DD3d_accumulate), rect1,
                                          rect2)

    def show_plots(self):
        """
        Show all the plots on the screen.

        This method displays all the plots generated by the class on the screen.
        If the `intermediate_plot` flag is set to True, the plots will be shown.
        """
        if self.intermediate_plot is True:
            self.plt.show()
